require('./index.css');


//GLOBAL
let tables = {};

const icons = {
    edit: require('./icons/edit.svg'),
    unsorted: require('./icons/unsorted-icon.svg'),
    sortAscending: require('./icons/sort-ascending-icon.svg'),
    sortDescending: require('./icons/sort-descending-icon.svg'),
}

class TableData
{
    constructor()
    {
        /** @type {Array<String>} */
        this.headers = [];

        /** @type {Array<String>} */
        this.dataTypes = [];

        /** @type {Array<Array<any>>} */
        this.rows = [];
    }
}

class Table
{
    constructor(el, data)
    {
        /** @type {HTMLDivElement} */
        this.el = el;

        /** @type {TableData} */
        this.data = data;

        /** @type {Array<Row>} */
        this.rows = [];

        /** @type {Row} */
        this.headerRow = null;
    }

    /**
     * 
     * @param {Row} row 
     */
    addRow(row)
    {
        this.el.appendChild(row.el);
        this.rows.push(row);
        row.table = this;
    }
}

class Row
{
    constructor()
    {
        /** @type {HTMLDivElement} */
        this.el = null;

        /** @type {Table} */
        this.table = null;

        /** @type {Array<Cell>} */
        this.cells = [];

        /** @type {Number} */
        this.rowId = null;
    }

    /**
     * 
     * @param {Number} rowId 
     */
    static make(rowId)
    {
        const el = newEl('div', {
            classes: ['row'],
            attributes: {
                rowId,
            }
        });

        const row = new Row();
        row.el = el;
        row.rowId = rowId;
        
        return row;
    }

    /**
     * 
     * @param {Cell} cell 
     */
    addCell(cell)
    {
        this.el.appendChild(cell.el);
        this.cells.push(cell);
        cell.row = this;
    }
}

class HeaderRow extends Row
{
    static make()
    {
        const el = newEl('div', {
            classes: ['header-row'],
            attributes: {
                rowId: 0
            }
        });

        const row = new Row();
        row.el = el;
        row.rowId = 0;
        
        return row;
    }

    /**
     * 
     * @param {HeaderCell} cell 
     */
    addCell(cell){ super.addCell(cell); }
}

class HeaderCell
{
    constructor()
    {
        /** @type {HTMLSpanElement} */
        this.el = null;

        /** @type {Row} */
        this.row = null;

        /** @type {HTMLImageElement} */
        this.sortButton = null;

        /** @type {Number} */
        this.colId = null;
    }
    
    static make(title, colId = null)
    {
        let el;
        if(colId != null)
        {
            el = newEl('span', {
                attributes: {
                    colId
                },
                text: title,
            });
        }
        else
        {
            el = newEl('span', {
                text: title,
            });
        }

        const headerCell = new HeaderCell();
        headerCell.el = el;
        headerCell.colId = colId;

        headerCell.sortButton = newEl('img', {
            attributes: {
                src: icons.unsorted,
                state: 'unsorted',
            },
            onclick: onSortButtonClick(headerCell),
            parent: el,
        });

        return headerCell;
    }
}

class Cell
{
    constructor()
    {
        /** @type {HTMLSpanElement} */
        this.el = null;

        /** @type {HTMLSpanElement} */
        this.dataSpan = null;

        /** @type {HTMLImageElement} */
        this.editButton = null;

        /** @type {Row} */
        this.row = null;

        /** @type {any} */
        this.data = {};

        /** @type {HTMLInputElement} */
        this.editField = null;

        /** @type {Boolean} */
        this.isEditing = false;

        /** @type {Number} */
        this.colId = null;
    }

    get parentTable()
    {
        return this.row.table;
    }

    /**
     * 
     */
    static make(colId, data = null)
    {
        const el = newEl('span', {
            classes: ['cell'],
            attributes: {
                colId
            }
        });

        const dataSpan = newEl('span', {
            parent: el,
            text: data.toString()
        });

        const editButton = newEl('img', {
            classes: ['edit-button'],
            parent: el,
            attributes: {
                src: icons.edit,
            }
        });

        const cell = new Cell();
        cell.el = el;
        cell.dataSpan = dataSpan;
        cell.editButton = editButton;
        cell.colId = colId;
        cell.data = data;


        editButton.onclick = onEditButtonClick(cell);


        return cell;
    }

    /**
     * 
     * @param {String} data 
     */
    setData(data)
    {
        this.dataSpan.innerText = data;
        this.data = data;
    }

    startEdit()
    {
        if(this.isEditing) return;

        this.dataSpan.style.display = 'none';

        const field = newEl('input', {
            classes: ['edit-field'],
            parent: this.el, 
        });

        field.value = this.data;
        field.onkeyup = onFieldSubmit(this);

        this.editField = field;
        this.isEditing = true;
    }

    /**
     * 
     * @param {Boolean} save 
     */
    stopEdit(save)
    {
        if(!this.isEditing) return;

        this.dataSpan.style.display = 'block';

        if(save)
        {
            this.setData(this.editField.value);
        }

        this.el.removeChild(this.editField);
        this.isEditing = false;
    }
}

function newEl(tag, options)
{
    /** @type {HTMLElement} */
    const el = document.createElement(tag);
    
    if(options.parent) options.parent.appendChild(el);

    editEl(el, options);
    
    return el;
}

function editEl(el, {id, parent, classes, attributes, text, onclick} = {})
{
    if(id) el.id = id;

    if(classes) el.classList.add(...classes);

    if(attributes)
    {
        for(const attr in attributes)
        {
            el.setAttribute(attr, attributes[attr]);
        }
    }

    if(text) el.innerText = text;
    if(onclick) el.onclick = onclick;
    
    return el;
}

function onEditButtonClick(cell)
{
    return () => cell.startEdit();
}

function onFieldSubmit(cell)
{
    return (e) =>
    {
        if(e.keyCode == 13) cell.stopEdit(true); //13 is the code for the Enter key
    }
}

/**
 * 
 * @param {Table} table 
 */
function generateRows(table)
{
    const data = table.data;

    const headerRow = HeaderRow.make()
    table.addRow(headerRow);
    table.headerRow = headerRow;

    for(const [i, title] of data.headers.entries())
    {
        headerRow.addCell(HeaderCell.make(title, i));
    }

    for(let i = 0; i < data.rows.length; i++)
    {
        const rowData = data.rows[i];

        const row = Row.make(i+1); //0 is the header row
        table.addRow(row);

        for(let i = 0; i < rowData.length; i++)
        {
            row.addCell(Cell.make(i, rowData[i]));
        }
    }
}


/**
 * 
 * @param {Table} table 
 * @param {Number} colIndex 
 */
function getSingleColumnMinWidth(table, colIndex)
{
    let minWidth = 0;

    for(const row of table.rows)
    {
        const colCellWidth = row.cells[colIndex].el.getBoundingClientRect().width;
        if(minWidth < colCellWidth)
        {
            minWidth = colCellWidth;
        }
    }

    return minWidth;
}

/**
 * 
 * @param {Table} table 
 */
function setMinWidths(table)
{
    const minWidths = [];

    for(const row of table.rows)
    {
        for(const [i, cell] of row.cells.entries())
        {
            if(minWidths.length - 1 < i)
            {
                minWidths.push(0);
            }

            const cellWidth = cell.el.getBoundingClientRect().width;
            if(minWidths[i] < cellWidth)
            {
                minWidths[i] = cellWidth;
            }
        }
    }

    for(const row of table.rows)
    {
        for(const [i, cell] of row.cells.entries())
        {
            cell.el.style.width = minWidths[i];
        }
    }
}

/**
 * 
 * @param {Table} table 
 */
function setHeaderPaddingOffset(table)
{
    table.el.parentElement.style.paddingTop = table.rows[0].el.getBoundingClientRect().height;
}

function tick()
{
    for(const tableId in tables)
    {
        const {headerRow, el} = tables[tableId];

        headerRow.el.style.width = el.offsetWidth + 'px';
        headerRow.el.scrollLeft = el.scrollLeft;
        
    }
}

/**
 * 
 * @param {Table} table 
 * @param {Number} colId 
 * @param {String} order 
 */
function sortRows(table, colId, order)
{
    if(order == 'unsorted')
    {
        table.rows.sort((a, b) => 
        {
            if(a.rowId > b.rowId) return 1;
            if(a.rowId < b.rowId) return -1;
            if(a.rowId == b.rowId) return 0;
        });
    }
    else
    {
        multiplier = 1;
        if(order == 'descending')
        {
            multiplier = -1;
        }

        table.rows.sort((a, b) =>
        {
            if(a.cells[colId].data > b.cells[colId].data) return 1 * multiplier;
            if(a.cells[colId].data < b.cells[colId].data) return -1 * multiplier;
            if(a.cells[colId].data == b.cells[colId].data) return 0;
        });
    }

    for(const [i, row] of table.rows.entries())
    {
        table.el.appendChild(row.el); //We just append each row to the end of the table so that they'll get in order
    }

}

function onSortButtonClick(cell)
{
    return () => cycleSortState(cell);
}

/**
 * 
 * @param {HeaderCell} cell 
 */
function cycleSortState(cell)
{
    const colId = cell.el.getAttribute('colId');
    let order = cell.sortButton.getAttribute('state');

    if(order == 'unsorted')
    {
        order = 'ascending';
        editEl(cell.sortButton, {
            attributes: {
                src: icons.sortAscending,
                state: 'ascending',
            },
        });
    }
    else if(order == 'ascending')
    {
        order = 'descending';
        editEl(cell.sortButton, {
            attributes: {
                src: icons.sortDescending,
                state: 'descending'
            },
        });
    }
    else
    {
        order = 'unsorted';
        editEl(cell.sortButton, {
            attributes: {
                src: icons.unsorted,
                state: 'unsorted'
            },
        });
    }

    sortRows(cell.row.table, colId, order);

    //set all other buttons as unordered
    for(const c of cell.row.cells)
    {
        if(c != cell)
        {
            editEl(c.el, {
                attributes: {
                    state: 'unsorted',
                    src: icons.unsorted,
                }
            });
        }
    }
}

function createTable(parentId, data)
{
    const tableId = 'table-' + parentId;

    const parent = document.getElementById(parentId);
    parent.classList.add('table-wrapper');

    const tableDiv = newEl('div', {
        id: tableId,
        classes: ['table'],
        parent: parent
    });

    tables[tableId] = new Table(tableDiv, data);

    generateRows(tables[tableId]);
    setMinWidths(tables[tableId]);
    setHeaderPaddingOffset(tables[tableId]);
}

function main()
{
    setInterval(tick, 1000/60);
}

main();

module.exports.createTable = createTable;
module.exports.TableData = TableData;
module.exports.Table = Table;
module.exports.Row = Row;
module.exports.HeaderRow = HeaderRow;
module.exports.Cell = Cell;
module.exports.HeaderCell = HeaderCell;